/* Clique */

#include <time.h>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>

using namespace std;

typedef unsigned int uint;

// this 3 tiny functions are defined below after the 'main' function
void to_upper_case ( vector<string> & seqs );
void u_to_t        ( vector<string> & seqs );
uint word_dist     ( const string & w1, const string & w2,
                     uint offset1, uint offset2, uint length );

// converts clock_t value to seconds
inline double clock2sec( clock_t t )
{
  return int(double(t)/CLOCKS_PER_SEC*100+.5)/100;
}

int main( int argc, char * argv[] )
{

  const char * program_name = "clique";     // program nema
  const char * logfile_name = "clique.log"; // log file name

  // saying hello
  cerr << program_name << ": Program for Conservative Sights Search\n"
       << "(C) 2006, Laboratory of Mathematical Methods and Models in Bioinformatics\n"
       << "Institute for Information Transmission Problems, Russian Academy of Sciences\n\n";

  // checking number of arguments in command line
  if ( argc <= 5 )
  {
    cerr << "Usage: " << endl
         << "  " << program_name << " in sl dm q out [-i|-u|-iu]" << endl
         << "where:" << endl
         << "  in  - input file name (string)"   << endl
         << "  sl  - signal length (integer)"    << endl
         << "  dm  - distance maximum (integer)" << endl
         << "  q   - clique size (integer)"      << endl
         << "  out - output file name (string)"  << endl
         << "  -i  - ignore case in sequences"   << endl
         << "  -u  - treat 'U' as if it was 'T'" << endl
         << "  -iu - turn on both switches"      << endl
         << "Note: arguments order is fixed.";
    return 1;
  }

  // opening log file
  ofstream logfile( logfile_name );
  if ( !logfile.is_open() ) // file opening is unsuccessful
  {
    cerr << "Couldn't open log file '" << logfile_name << "'." << endl;
    return 1;
  }
  logfile << program_name << " program log file." << endl << endl;

  // processing command line parameters
  const char * infile_name = argv[1]; // input file name
  const uint sl = atoi(argv[2]); // signal length
  if ( !sl )
  {
    cerr << "Incorrect value for signal length: '" << argv[2] << "'." << endl;
    return 1;
  }
  const uint dm = atoi(argv[3]); // distance maximum
  if ( !dm && !( argv[3][0]=='0' && argv[3][1]=='\0' )  )
  {
    cerr << "Incorrect value for distance maximum: '" << argv[3] << "'." << endl;
    return 1;
  }
  const uint q  = atoi(argv[4]); // clique size
  if ( !q )
  {
    cerr << "Incorrect value for clique size: '" << argv[4] << "'." << endl;
    return 1;
  }
  const char * outfile_name = argv[5]; // output file name
  bool ignore_case = false;
  bool identify_ut = false;
  if ( argc == 7 )
  {
    string sw = argv[6];
    if ( sw == "-i" )
      ignore_case = true;
    else if ( sw == "-u" )
      identify_ut = true;
    else if ( sw == "-iu" )
      ignore_case = true,
      identify_ut = true;
    else
      cerr << "Unknown command line parameter '" << sw << "' ignored." << endl << endl;
      
  }

  // sending program parameters to log file
  logfile << "Program parameters:" << endl;
  logfile << "in  = " << infile_name << endl;
  logfile << "sl  = " << sl << endl;
  logfile << "dm  = " << dm << endl;
  logfile << "q   = " << q  << endl;
  logfile << "out = " << outfile_name << endl;
  if ( ignore_case )
    logfile << "Ignoring case in input sequences." << endl;
  if ( identify_ut )
    logfile << "Identifying U and T in input sequences." << endl;
  logfile << endl;

  logfile << "Reading sequences from infile." << endl;

  // getting sequences from input file
  vector<string> sids;  // sequences identifiers
  vector<string> seqs;  // sequences themselves
  vector<string> seqs_; // sequences copy for comparison
  ifstream infile( infile_name ); // opening input file
  if ( !infile.is_open() ) // file opening is unsuccessful
  {
    cerr << "Couldn't open input file '" << infile_name << "'." << endl;
    return 1;
  }
  while ( !infile.eof() ) // reading input file line by line
  {
    string line; // current line
    getline( infile, line ); // reading line from file
    if ( line.empty() ) // empty line
      continue; // skipping
    if ( line[0] == '>' ) // sequence identifier (header)
    {
      // adding sequence identifier to the sids vector
      sids.push_back( line.substr( 1, line.length()-1 ) );
      // adding empty sequence body to the seqs vector
      seqs.push_back( string() );
      // and also a copy for future case conversion etc.
      seqs_.push_back( string() );
    }
    else // regular sequence line
      if ( !seqs.empty() ) // at least one sequence identifier has been read yet
        seqs[seqs.size()-1] += line,   // adding line to current sequence in vector
        seqs_[seqs_.size()-1] += line; // and a copy ...
  }
  infile.close(); // closing input file
  const uint numseqs = seqs.size(); // number of sequences

  // sending sequences to log
  logfile << "Successfully read " << numseqs << " sequences from infile:" << endl;
  for ( uint i=0; i<numseqs; i++ )
    logfile << '>' << sids[i] << endl
            << seqs[i] << endl;
  logfile << endl;

  // preparing sequences for distance calculation
  if ( ignore_case )
    to_upper_case(seqs_); // converting all sequences (copies) to upper case
  if ( identify_ut )
    u_to_t(seqs_); // replace all 'U' letters by 'T'

  // checking and caching sequence lengths
  uint numverts = 0; // number of vertices in graph
  // lengths (sizes) of the graph parts
  // (part = subset of vertices corresponding to one sequence)
  uint * partlens = new uint[numseqs];
  // offsets of parts in the vertices list
  uint * partbegs = new uint[numseqs+1];
  partbegs[0] = 0;
  for ( uint i=0; i<numseqs; i++ )
  {
    if ( seqs[i].size() < sl )
    {
      cerr << "Sequence No." << i << " ('" << sids[i]
           << "') is shorter than the cpecified signal length ("
           << seqs[i].size() << "<" << sl << ")." << endl;
      return 1;
    }
    partlens[i] = seqs[i].size() - sl + 1;
    numverts += partlens[i];
    partbegs[i+1] = numverts;
  }
  logfile << "Number of vertices in graph = " << numverts << "." << endl;

  // allocating adjacency matrix
  logfile << "Allocating adjacency matrix ("
          << numverts << "x" << numverts << ")." << endl;
  bool ** matrix=0;   // adjacency matrix
  uint *  vert_deg=0; // degree of vertices
  uint *  vert2seq=0; // vertex to sequence correspondence
  bool *  vert_NiC=0; // vertex is not in *isolated* clique
  try
  {
    matrix   = new bool*[numverts];
    vert_deg = new uint[numverts];
    vert2seq = new uint[numverts];
    vert_NiC = new bool[numverts];
    for ( uint i=0; i<numverts; i++ )
    {
      matrix[i] = new bool[numverts];
      for ( uint j=0; j<numverts; j++ )
        matrix[i][j] = false;
      vert_deg[i] = 0;
      vert2seq[i] = 0;
      vert_NiC[i] = false;
    }
  }
  catch ( bad_alloc )
  {
    cerr << "Not enough memory to allocate adjacency matrix." << endl;
    return 1;
  }

  // caching vertex to sequence correspondence
  for ( uint k=0; k<numseqs; k++ )
    for ( uint i=0; i<partlens[k]; i++ )
      vert2seq[partbegs[k]+i] = k;

  // calculating distances, bulding the graph
  cerr << "Building the graph ... ";
  logfile << "Calculating distances, bulding the graph." << endl;
  uint numedges = 0; // number of edges in graph
  for ( uint k=1; k<numseqs; k++ )
    for ( uint l=0; l<k; l++ )
      for ( uint i=0; i<partlens[k]; i++ )
        for ( uint j=0; j<partlens[l]; j++ )
          if ( word_dist( seqs_[k], seqs_[l], i, j, sl ) <= dm )
            matrix[partbegs[k]+i][partbegs[l]+j] = true,
            matrix[partbegs[l]+j][partbegs[k]+i] = true,
            vert_deg[partbegs[k]+i]++,
            vert_deg[partbegs[l]+j]++,
            numedges+=2;
  logfile << "Finished building the graph." << endl;
  logfile << "Number of edges in graph = " << numedges << "." << endl;
  cerr << "OK" << endl;

  // opening output file
  ofstream outfile( outfile_name ); // opening output file
  if ( !outfile.is_open() ) // file opening is unsuccessful
  {
    cerr << "Couldn't open output file '" << outfile_name << "'." << endl;
    return 1;
  }

  // reducing the graph and searching for isolated cliques
  cerr << "Reducing the graph and searching for isolated cliques ... ";
  clock_t clock_calc = clock();
  clock_t clock_tri  = 0;
  clock_t clock_tetr = 0;
  do
  {
    uint pass_n=0;
    uint numedges0;
    uint i_min=0, j_min=0; // edge (i,j) belongs to the minimal number of triangles
    uint min_tri = 0; // minimal number of triangles
    do
    {
      pass_n++;
      numedges0 = numedges;
  
      // removing all edges from vertices connected to less than q-1 graph parts
      for ( uint i=0; i<numverts; i++ )
        if ( vert_deg[i] )
        {
          uint numlinks = 0; // number of links to other parts of graph
          for ( uint k=0; k<numseqs; k++ )
            for ( uint j=0; j < partlens[k]; j++ )
              if ( matrix[i][partbegs[k]+j] )
              {
                numlinks++;
                break;
              }
          if ( numlinks < q-1 )
          {
            for ( uint j=0; j<numverts; j++ ) // removing all edges from the vertex i
              if ( matrix[i][j] )
                matrix[i][j] = false,
                matrix[j][i] = false,
                vert_deg[j]--,
                numedges -= 2;
            vert_deg[i] = 0;
          }
        }
  
      // removing all edges belonging to less than q-2 triangles
      i_min = j_min = min_tri = 0;
      if ( ! (numedges < numedges0) )
      {
        clock_t tick = clock();
        logfile << "Counting triangles..." << flush;
        for ( uint i=0; i<numverts; i++ )
          if ( vert_deg[i] )
            for ( uint j=i+1; j<numverts; j++ )
              if ( matrix[i][j] )
              {
                uint a_sq=0; // (matrix^2)[i][j]
                for ( uint k=0; k<numverts; k++ )
                  if ( matrix[i][k] && matrix[k][j] )
                    a_sq++;
                if ( a_sq < q-2 )
                  matrix[i][j] = false,
                  matrix[j][i] = false,
                  vert_deg[i]--,
                  vert_deg[j]--,
                  numedges -= 2;
                else
                  if ( a_sq < min_tri || !min_tri )
                    min_tri = a_sq,
                    i_min = i,
                    j_min = j;
              }
        tick = clock()-tick;
        clock_tri += tick;
        logfile << "done (" << clock2sec(tick) << "s)." << endl;
      }
  
      // removing all edges belonging to less than (q-2)(q-3)/2 tetrahedrons
      if ( ! (numedges < numedges0) )
      {
        clock_t tick = clock();
        logfile << "Counting tetrahedrons... " << flush;
        for ( uint i=0; i<numverts; i++ )
          if ( vert_deg[i] )
            for ( uint j=i+1; j<numverts; j++ )
              if ( matrix[i][j] )
              {
                uint num_tetr = 0;
                for ( uint k=0; k<numverts; k++ )
                  if ( k!=i && k!=j && vert_deg[k] )
                    for ( uint l=k+1; l<numverts; l++ )
                      if ( l!=i && l!=j && l!=k && matrix[k][l] )
                        if ( matrix[i][k] && matrix[i][l] &&
                             matrix[j][k] && matrix[j][l] )
                          num_tetr++;
                if ( num_tetr < (q-2)*(q-3)/2 )
                  matrix[i][j] = false,
                  matrix[j][i] = false,
                  vert_deg[i]--,
                  vert_deg[j]--,
                  numedges -= 2;
              }
        tick = clock()-tick;
        clock_tetr += tick;
        logfile << "done (" << clock2sec(tick) << "s)." << endl;
      }
  
      logfile << "Pass " << pass_n << ": removed "
              << (numedges0-numedges) << " edges." << endl;
  
    }
    while ( numedges < numedges0 );
    logfile << "Number of edges in graph = " << numedges << "." << endl;

    // checking that there are still edges in the graph  
    if ( numedges == 0 )
    {
      logfile << "Graph is empty." << endl;
      break;
    }

    // marking as "not in clique" all vertices with degree not equal to q-1
    for ( uint i=0; i<numverts; i++ )
      vert_NiC[i] = ( vert_deg[i] != q-1 );

    // checking if there is an isolated clique in the graph
    bool found_clique = false; // "found a clique" flag
    uint clique_n = 0; // number of the last found clique
    for ( uint i=0; i<numverts; i++ ) // for each vertex i ...
    {
      if ( !vert_NiC[i] ) // ... that is not marked as "not in clique"
        for ( uint j=0; j<numverts && !vert_NiC[i]; j++ ) // for each vertex j ...
          if ( matrix[i][j] ) // ... incident to vertex i
            if ( vert_NiC[j] ) // if j is marked as "not in clique"
              vert_NiC[i] = true; // mark vortex i "not in clique"
            else // else (j is unmarked (and has degree equal to q-1))
              for ( uint k=0; k<numverts; k++ ) // for each vertex k ...
                if ( k!=i && k!=j ) // ... not equal to i nor j ...
                  if ( matrix[i][k] && !matrix[j][k] ) // ... incident to i but not to j
                  {
                    vert_NiC[i] = true,
                    vert_NiC[j] = true; // mark both i and j as "not in clique"
                    break;
                  }
      if ( !vert_NiC[i] ) // if i is still unmarked => it's the 1st vertex of a clique
      {
        found_clique = true; // FOUND CLIQUE!
        logfile << "Found a clique." << endl;

        vector<int> sns; // vector of sequence numbers ...
        vector<int> sos; // ... and corresponding offsets

        // getting coordinates of sequence fragments corresponding to clique
        for ( uint j=0; j<numverts; j++ )
          if ( j==i || matrix[i][j] )
            sns.push_back( vert2seq[j] ), // sequence number
            sos.push_back( j - partbegs[vert2seq[j]] ); // offset in sequence

        // counting absolutely conservative positions
        uint abs_cons_n = sl;
        for ( uint p=0; p<sl; p++ ) // position (0<=p<sl)
          for ( uint n=1; n<q; n++ )  // sequence fragment number (1<=n<q)
            if ( seqs_[sns[n]][sos[n]+p] != seqs_[sns[0]][sos[0]+p] )
            {
              abs_cons_n--;
              break;
            }

        // sending clique to output file
        outfile << ">clique" << (++clique_n) << " " << abs_cons_n << endl;
        for ( uint n=0; n<q; n++ )
          outfile << seqs[sns[n]].substr(sos[n],sl) << "\t"
                  << sids[sns[n]] << ", " << sos[n] << endl;
        outfile << endl;

        // removing clique from the graph
        for ( uint j=0; j<numverts; j++ )
          if ( j==i || matrix[i][j] )
          {
            for ( uint k=j; k<numverts; k++ )
              if ( matrix[i][j] )
                matrix[j][k] = false,
                matrix[k][j] = false;
            vert_deg[j] = 0;
            vert_NiC[j] = true;
          }
        numedges -= (q-1)*q;
      }
    }

    // remove 1 edge belonging to the minimal number of triangles
    if ( !found_clique )
      matrix[i_min][j_min] = false,
      matrix[j_min][i_min] = false,
      vert_deg[i_min]--,
      vert_deg[j_min]--,
      vert_NiC[i_min] = false,
      vert_NiC[j_min] = false,
      numedges -= 2,
      min_tri = 0,
      i_min = 0,
      j_min = 0;

  }
  while ( numedges > 0 );
  cerr << "OK" << endl;
  outfile.close(); // closing output file
  clock_calc = clock()-clock_calc;
  logfile << "Calculations complete." << endl
          << "Total calc. time: "
          << clock2sec(clock_calc) << "s including "
          << clock2sec(clock_tri)  << "s spent on counting triangles and "
          << clock2sec(clock_tetr) << "s spent on counting tetrahedrons." << endl;

  // deallocating adjacency matrix
  cerr << "Releasing memory ... ";
  logfile << "Deallocating adjacency matrix." << endl;
  for ( uint i=0; i<numverts; i++ )
    delete[] matrix[i];
  delete[] matrix;
  delete[] vert_deg;
  delete[] vert2seq;
  delete[] vert_NiC;
  delete[] partlens;
  delete[] partbegs;
  cerr << "OK" << endl;

  cerr << endl << "Thank you for using the program." << endl;

  return 0;
}

// convert all characters' case of all strings in vector to upper case
void to_upper_case( vector<string> & seqs )
{
  for ( uint i=0; i<seqs.size(); i++ )
    for ( uint j=0; j<seqs[i].size(); j++ )
      if ( seqs[i][j] >= 'a' && seqs[i][j] <= 'z' )
        seqs[i][j] = seqs[i][j] - 'a' + 'A';
}

// replace all 'U' characters in all strings in vector by 'T'
void u_to_t( vector<string> & seqs )
{
  for ( uint i=0; i<seqs.size(); i++ )
    for ( uint j=0; j<seqs[i].size(); j++ )
      if ( seqs[i][j] == 'U' )
        seqs[i][j] = 'T';
      else if ( seqs[i][j] == 'u' )
        seqs[i][j] = 't';
}

// Calculate distance between two words
// specified by their coordinates in two given sequences
// and their (common) length.
uint word_dist( const string & w1, const string & w2,
                uint x1, uint x2, uint length )
{
  uint dist=0;
  for ( uint i=0; i<length; i++ )
    if ( w1[x1+i] != w2[x2+i] )
      dist++;
  return dist;
}
